   /* AverAgeDepBP.c */
   /* Averaged age-dependent branching process with Type I,II,or III life time:
      Distribution of the process */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

double frand(void)
{
 return 1.0-(double)rand()/(RAND_MAX+1.0);
}

#define maxsize 30001   /* Maximum population size */

double mean(p)
 double p[5];
 {
  double mu=0.0;
  int i;
  for(i=1;i<5;++i) mu +=i*p[i];
  return mu;
 }
int offspring(p,v)
 double p[5],v;
 {
  int i=0;
  double u=p[0];
  while ((i<4)&&(v>=u)) {
   ++i;u+=p[i];
  }
  return i;
 }
float minlist(L)
 float L[maxsize];
 {
  long int i=1;
  float a=L[0];
  while (i<maxsize && L[i]>0.0) {
   if (a>L[i]) a=L[i];
   ++i;
  }
  return a;
 }
void renew(L,children,lives)
 float L[maxsize],lives[4];
 int children;
 {
  int k;
  long int i=1,j=0;
  float a=L[0];
  while (i<maxsize && L[i]>0.0) {
   if (a>L[i]) {a=L[i];j=i;}
   ++i;
  }
  if (i<maxsize-children+1) {
   if (children>1) {
    while (i>j) {
     L[i-1+children]=L[i];
     --i;
    }
   }
   if (children>0) {
    for (k=0;k<children;k++) L[j+k]=a+lives[k];
   }
   else {
    while (L[j]>0.0) {
     L[j]=L[j+1];
     ++j;
    }
   }
  }
 }

main(void)
{
 int j,k,n,
     sample, /* number of repetition of simulations */
     init; /* initial size */
 long int i,seed,
          x,y; /* size after and before each time of the process */
 unsigned long int s[10001];
 float a, /* exponent of life time distribution */
       simtime, /* time to be simulated */
       L[maxsize], /* list of life times of a current population */
       lives[4]; /* list of life times of children */
 double p[5], /* probability of giving birth to babies */
        d, /* atomic time for sum =simtime/10000 */
        t; /* time when birth or death occurs */
 char *file1 = "datax.dat",*file2 = "distx.dat",*file3 = "averagex.dat";
 FILE *fp1,*fp2,*fp3;

   /*----------------- Input initial parameters -----------------*/

 printf("Exponent a of life time dist.(a>1 for type I, a=1 for type II, and 0<a<1 for type III)=");
 scanf("%f",&a);
 printf("\n");
 printf(" Offspring distribution:\n\t");
 for(i=0;i<4;++i){
  printf("p(%d)=",i);
  scanf("%lf",&p[i]);
  printf("\t");
 }
 p[4]=1.0-p[0]-p[1]-p[2]-p[3];
 printf("\n");
 printf(" Mean offspring number = %4.3f\n",mean(p));
 printf("\n");
 printf(" Time to be observed (positive number) = ");
 scanf("%f",&simtime);
 printf("\n");
 d=simtime/10000.0;
 printf(" Initial population size (any positive integer) = ");
 scanf("%d",&init);
 printf("\n");
 printf(" Samples(<10000) = ");
 scanf("%d",&sample);
 printf("\n");
 printf(" Random number seed (any integer) = ");
 scanf("%d",&seed);
 printf("\n\n");

  /* ---- Print initial parameters in a file "datax.dat" ----*/

 fp1=fopen(file1,"w");
 fprintf(fp1," << Averaged age-dependent branching process with Type I,II, or III life times >>\n");
 fprintf(fp1,"  < Distributions of the process >\n\n");
 fprintf(fp1," Exponent of life time dist. = %4.2f\n\n",a);
 fprintf(fp1," Mean life time = %4.2f\n\n",a/(a+1.0));
 fprintf(fp1," Offspring distribution:\n");
 for(i=0;i<5;++i) fprintf(fp1,"  p(%d)=%3.3f  ",i,p[i]);
 fprintf(fp1,"\n\n");
 fprintf(fp1," Mean offspring number = %4.3f\n\n",mean(p));
 fprintf(fp1," Time to be observed = %4.2f\n\n",simtime);
 fprintf(fp1," Unit of time = %6.5f\n\n",d);
 fprintf(fp1," Initial population size = %d\n\n",init);
 fprintf(fp1," Samples = %d\n\n",sample);
 fprintf(fp1," Random number seed = %d\n",seed);
 fclose(fp1);

  /* Start simulation to save the results in files "distx.dat" and  
  "averagex.dat"*/

 fp2=fopen(file2,"w");
 srand(seed); /* giving the seed of frand */
 for (n=1;n<10001;++n) s[n]=0;
 for (j=1;j<=sample;++j) {
  for (i=0;i<maxsize;++i) L[i]=(float)0.0;
  for (i=0;i<init;++i) L[i]=(float)pow(frand(),1.0/a);
  for (i=0;i<4;++i) lives[i]=(float)0.0;
  x=init;t=0.0;n=1;
  while (t<simtime && x!=0 && x<maxsize) {
   t=minlist(L);
   k=offspring(p,frand());
   for (i=0;i<k;++i) lives[i]=(float)pow(frand(),1.0/a);
   y=x;x+=k-1;
   while (n<10001 && n*d<t) {s[n]+=y;++n;}
   if (x==0) y=0;
   renew(L,k,lives);
  }
  printf("%8d:%7d",j,y);
  fprintf(fp2,"%d\n",y);
 }
 fclose(fp2);
 fp3=fopen(file3,"w");
 fprintf(fp3,"0.0\t %d\n",init);
 for(n=1;n<10001;++n)
  fprintf(fp3,"%5.3f\t %f\n",(double)n*d,(double)s[n]/(double)sample);
 fclose(fp3);
}
