   /* GeneralCBP.c */
   /* General branching process with Type I,II,or III life time :
      Age distribution and the distribution of life remaining */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

double frand(void)
{
 return 1.0-(double)rand()/(RAND_MAX+1.0);
}

#define maxsize 30001   /* Maximum population size */

double mean(p)
 double p[5];
 {
  double mu=0.0;
  int i;
  for(i=1;i<5;++i) mu +=i*p[i];
  return mu;
 }
int offspring(p,v)
 double p[5],v;
 {
  int i=0;
  double u=p[0];
  while ((i<4)&&(v>=u)) {
   ++i;u+=p[i];
  }
  return i;
 }
float minlist1(BDL)
 float BDL[maxsize][3];
 {
  long int i=1;
  float a=BDL[0][1];
  while (i<maxsize && BDL[i][1]>0.0) {
   if (a>BDL[i][1]) a=BDL[i][1];
   ++i;
  }
  return a;
 }
float minlist2(BDL)
 float BDL[maxsize][3];
 {
  long int i=1;
  float a=BDL[0][2];
  while (i<maxsize && BDL[i][2]>0.0) {
   if (a>BDL[i][2]) a=BDL[i][2];
   ++i;
  }
  return a;
 }
void renew1(BDL,children,birth,lives)
 float BDL[maxsize][3],birth[4],lives[4];
 int children;
 {
  int k;
  long int i=1,j=0;
  float a=BDL[0][1];
  while (i<maxsize && BDL[i][2]<100000.0) {
   if (a>BDL[i][1]) {a=BDL[i][1];j=i;}
   ++i;
  }
  BDL[j][1]=(float)100000.0;
  if (i<maxsize-children+1 && children>0) {
   while (i>j+1) {
    for (k=0;k<3;k++) BDL[i-1+children][k]=BDL[i-1][k];
    --i;
   }
   for (k=1;k<children+1;k++) {
    BDL[j+k][0]=a;BDL[j+k][1]=a+birth[k-1];BDL[j+k][2]=a+lives[k-1];
   }
  }
 }
void renew2(BDL)
 float BDL[maxsize][3];
 {
  int k;
  long int i=1,j=0;
  float a=BDL[0][2];
  while (i<maxsize && BDL[i][2]<100000.0) {
   if (a>BDL[i][2]) {a=BDL[i][2];j=i;}
   ++i;
  }
  while (BDL[j][2]<100000.0) {
   for (k=0;k<3;k++) BDL[j][k]=BDL[j+1][k];
   ++j;
  }
 }

main(void)
{
 int j,k;
 long int i,seed,
          x; /* size at each time of the process */
 float a, /* exponent of life time distribution */
       simtime, /* time to be simulated */
       BDL[maxsize][3], /* list of birth, death, and childbirth times of
a current population */
       birth[4], /* list of childbirth times */
       lives[4], /* list of life times of children */
       t; /* time when birth or death occurs */
  double p[5]; /* probability of giving birth to babies */
  char *file1 = "datax.dat",*file2 = "samplex.dat",
       *file3 = "liferemainx.dat",*file4 = "agex.dat";
  FILE *fp1,*fp2,*fp3,*fp4;

   /*----------------- Input initial parameters -----------------*/

 printf("Exponent a of life time dist.(a>1 for type I, a=1 for type II, and 0<a<1 for type III)=");
 scanf("%f",&a);
 printf("\n");
 printf(" Offspring distribution:\n\t");
 for(i=0;i<4;++i){
  printf("p(%d)=",i);
  scanf("%lf",&p[i]);
  printf("\t");
 }
 p[4]=1.0-p[0]-p[1]-p[2]-p[3];
 printf("\n");
 printf(" Mean offspring number = %4.3f\n",mean(p));
 printf("\n");
 printf(" Time to be observed (positive number) = ");
 scanf("%f",&simtime);
 printf("\n");
 printf(" Initial population size (any positive integer) = ");
 scanf("%d",&x);
 printf("\n");
 printf(" Random number seed (any integer) = ");
 scanf("%d",&seed);
 printf("\n\n");

  /* ------- Print initial parametersin a file "datax.dat" -------*/

 fp1=fopen(file1,"w");
 fprintf(fp1,"  << General branching process with Type I,II,or III life time>>\n");
 fprintf(fp1,"   < Age distribution and the distribution of life remaining> \n\n");
 fprintf(fp1," Exponent of life time dist. = %4.2f\n\n",a);
 fprintf(fp1," Mean life time = %4.2f\n\n",a/(a+1.0));
 fprintf(fp1," Time of childbirth is unif. distributed between 0.15 and 0.45\n\n");
 fprintf(fp1," Offspring distribution:\n");
 for(i=0;i<5;++i) fprintf(fp1,"  p(%d)=%3.3f  ",i,p[i]);
 fprintf(fp1,"\n\n");
 fprintf(fp1," Mean offspring number = %4.3f\n\n",mean(p));
 fprintf(fp1," Time to be observed = %4.2f\n\n",simtime);
 fprintf(fp1," Initial population size = %d\n\n",x);
 fprintf(fp1," Random number seed = %d\n",seed);
 fclose(fp1);

  /* Start simulation to save the results in files "samplex.dat",
  "liferemainx.dat" and "agex.dat" */

 fp2=fopen(file2,"w");
 srand(seed); /* giving the seed of frand */
 fprintf(fp2,"%5.3f\t %10d\n",0.0,x);
 for (i=0;i<maxsize;++i) {
  for (j=0;j<3;++j) BDL[i][j]=(float)100000.0;
 }
 for (i=0;i<x;++i) {
  BDL[i][0]=(float)0.0;BDL[i][2]=(float)pow(frand(),1.0/a);
  BDL[i][1]=(float)(0.2*(1.0+2.0*frand())*(double)BDL[i][2]);
  if (BDL[i][1]<0.15) BDL[i][1]=(float)100000.0;
 }
 for (j=0;j<4;++j) {birth[j]=(float)0.0;lives[j]=(float)0.0;}
 t=0.0;
 while ((t<simtime)&&(x!=0)&&(x<maxsize)) {
  if (minlist1(BDL)<minlist2(BDL)) {
   t=minlist1(BDL);
   k=offspring(p,frand());
   for (j=0;j<k;++j) {
    lives[j]=(float)pow(frand(),1.0/a);
    birth[j]=(float)(0.15*(1.0+2.0*frand()));
    if (birth[j]>lives[j]) birth[j]=(float)100000.0;
   }
   x+=k;
   fprintf(fp2,"%10.8f\t %10d\n",t,x);
   renew1(BDL,k,birth,lives);
  }
  else {
   t=minlist2(BDL);--x;
   fprintf(fp2,"%10.8f\t %10d\n",t,x);
   renew2(BDL);
  }
 }
 fclose(fp2);
 fp3=fopen(file3,"w");fp4=fopen(file4,"w");
 i=1;
 while (i<maxsize && BDL[i][2]<100000.0) {
  fprintf(fp3,"%f\n",BDL[i][2]-simtime);
  fprintf(fp4,"%f\n",simtime-BDL[i][0]);
  ++i;
 }
 fclose(fp3);fclose(fp4);
}
